@@ -44,6 +44,7 @@ from fluxrpc.transports.socket.messages import (
     SessionKeyMessage,
     TestMessage,
     LivelinessMessage,
+    AesRekeyMessage,
 )
 
 
@@ -57,15 +58,18 @@ def bytes_to_human(num, suffix="B"):
 
 @dataclass
 class KeyData:
-    rsa_key: str = ""
     aes_key: str = ""
-    private: str = ""
-    public: str = ""
+    rsa_private: str = ""
+    rsa_public: str = ""
+
+    def burn_rsa_keys(self):
+        self.rsa_private = ""
+        self.rsa_public = ""
 
     def generate(self):
-        self.rsa_key = RSA.generate(2048)
-        self.private = self.rsa_key.export_key()
-        self.public = self.rsa_key.publickey().export_key()
+        rsa_key = RSA.generate(2048)
+        self.rsa_private = rsa_key.export_key()
+        self.rsa_public = rsa_key.publickey().export_key()
 
 
 @dataclass
@@ -83,6 +87,9 @@ class EncryptablePeerGroup:
         peer: EncryptablePeer = self.get_peer(id)
         if peer:
             peer.read_socket_task.cancel()
+            for task in peer.in_flight_messages:
+                task.cancel()
+
             try:
                 peer.writer.close()
                 await peer.writer.wait_closed()
@@ -112,7 +119,7 @@ class EncryptablePeerGroup:
             print("destroy all peers exception")
             print(repr(e))
 
-    async def start_peer_timeout(self, id):
+    def start_peer_timeout(self, id):
         peer = self.get_peer(id)
         timeout = 10
 
@@ -135,6 +142,7 @@ class EncryptablePeer:
     proxied: bool = False
     timer: asyncio.Task | None = None
     read_socket_task: asyncio.Task | None = None
+    in_flight_messages: list[asyncio.Task] = field(default_factory=list)
     challenge_complete_event: asyncio.Event = field(default_factory=asyncio.Event)
     forwarding_event: asyncio.Event = field(default_factory=asyncio.Event)
     fh: dict[str, BinaryIO] = field(default_factory=dict)
@@ -250,24 +258,37 @@ class EncryptedSocketServerTransport(ServerTransport):
 
     def parse_session_key_message(self, key_pem: str, msg: SessionKeyMessage) -> str:
         """Used by Node to decrypt and return the AES Session key using the RSA Key"""
-        private_key = RSA.import_key(key_pem)
+        try:
+            private_key = RSA.import_key(key_pem)
 
-        enc_session_key = msg.rsa_encrypted_session_key
+            enc_session_key = msg.rsa_encrypted_session_key
 
-        cipher_rsa = PKCS1_OAEP.new(private_key)
-        session_key = cipher_rsa.decrypt(enc_session_key)
+            cipher_rsa = PKCS1_OAEP.new(private_key)
+            session_key = cipher_rsa.decrypt(enc_session_key)
 
-        enc_aes_key_message = SerializedMessage(msg.aes_key_message).deserialize()
+            enc_aes_key_message = SerializedMessage(msg.aes_key_message).deserialize()
+
+            aes_key_message = enc_aes_key_message.decrypt(session_key)
+        except Exception as e:
+            print(repr(e))
+            print("trower")
+            exit(0)
 
-        aes_key_message = enc_aes_key_message.decrypt(session_key)
         return aes_key_message.aes_key
 
-    async def begin_encryption(self, peer):
+    async def begin_encryption(self, peer: EncryptablePeer, rekey: bool = False):
         peer.key_data.generate()
-        msg = RsaPublicKeyMessage(peer.key_data.public)
+        msg = RsaPublicKeyMessage(peer.key_data.rsa_public)
 
-        await peer.send(msg.serialize())
-        await self.peers.start_peer_timeout(peer.id)
+        try:
+            if rekey:
+                msg = msg.encrypt(peer.key_data.aes_key)
+
+            await peer.send(msg.serialize())
+            self.peers.start_peer_timeout(peer.id)
+        except Exception as e:
+            print(repr(e))
+            exit("bye")
 
     async def send_challenge_message(self, peer: EncryptablePeer) -> bool:
         source = peer.writer.get_extra_info("sockname")
@@ -278,7 +299,7 @@ class EncryptedSocketServerTransport(ServerTransport):
             msg = self.auth_provider.generate_challenge(msg)
 
         await peer.send(msg.serialize())
-        await self.peers.start_peer_timeout(peer.id)
+        self.peers.start_peer_timeout(peer.id)
 
     async def valid_source_ip(self, peer_ip) -> bool:
         """Called when connection is established to verify correct source IP"""
@@ -325,6 +346,16 @@ class EncryptedSocketServerTransport(ServerTransport):
         reply = LivelinessMessage(msg.text[::-1])
         await peer.send(reply.serialize())
 
+    async def handle_aes_rekey_message(
+        self, peer: EncryptablePeer, msg: AesRekeyMessage
+    ):
+        # still figuring what to, is there any content in this message?
+        try:
+            await self.begin_encryption(peer, rekey=True)
+        except Exception as e:
+            print(repr(e))
+            exit("peace")
+
     async def handle_forwarding_message(self, peer: EncryptablePeer, msg: ProxyMessage):
         resp = ProxyResponseMessage(False)
         if msg.proxy_required:
@@ -348,20 +379,26 @@ class EncryptedSocketServerTransport(ServerTransport):
         await peer.send(resp.serialize())
         peer.forwarding_event.set()
 
-    async def handle_encryption_message(self, peer, msg):
-        peer.timer.cancel()
+    async def handle_encryption_message(
+        self, peer: EncryptablePeer, msg: SessionKeyMessage | EncryptedMessage
+    ):
+        if peer.timer:
+            peer.timer.cancel()
 
         if isinstance(msg, SessionKeyMessage):
-            aes_key = self.parse_session_key_message(peer.key_data.private, msg)
+            aes_key = self.parse_session_key_message(peer.key_data.rsa_private, msg)
 
             peer.key_data.aes_key = aes_key
 
-            # Send a test encryption request, always include random data
-            peer.random = get_random_bytes(16).hex()
-            test_msg = TestMessage(peer.random)
-            encrypted_test_msg = test_msg.encrypt(aes_key)
+            peer.key_data.burn_rsa_keys()
 
-            await peer.send(encrypted_test_msg.serialize())
+            if not peer.encrypted:
+                # Send a test encryption request, always include random data
+                peer.random = get_random_bytes(16).hex()
+                test_msg = TestMessage(peer.random)
+                encrypted_test_msg = test_msg.encrypt(aes_key)
+
+                await peer.send(encrypted_test_msg.serialize())
 
         if isinstance(msg, EncryptedMessage):
             response = msg.decrypt(peer.key_data.aes_key)
@@ -487,7 +524,8 @@ class EncryptedSocketServerTransport(ServerTransport):
 
         if peer.writer and not peer.writer.is_closing() and not peer.proxied:
             await self.begin_encryption(peer)
-        log.info("Handle client finished... waiting on read loop")
+
+        log.info("Client bootstrap complete... read loop has control")
 
     async def overrun_strategy(self, peer: EncryptablePeer, to_consume: int):
         log.info(f"LimitOverrun: Read {bytes_to_human(to_consume)} into buffer")
@@ -558,6 +596,7 @@ class EncryptedSocketServerTransport(ServerTransport):
         buffer = []
         while peer.reader and not peer.proxied and not peer.reader.at_eof():
             try:
+                # this needs timeout
                 data = await peer.reader.readuntil(separator=self.separator)
             except asyncio.exceptions.IncompleteReadError:
                 log.debug(f"Reader is at EOF. Peer: {peer.id}")
@@ -587,7 +626,12 @@ class EncryptedSocketServerTransport(ServerTransport):
                 buffer = []
 
             message = data.rstrip(self.separator)
-            await self.process_messages(peer, [message])
+
+            t = asyncio.create_task(
+                self.process_messages(peer, [message]), name=f"process_messages"
+            )
+
+            peer.in_flight_messages.append(t)
 
         log.debug(f"Read socket loop finished. peer.proxied: {peer.proxied}")
         if not peer.proxied:
@@ -608,41 +652,51 @@ class EncryptedSocketServerTransport(ServerTransport):
             log.debug(f"Received : {type(message).__name__}")
 
             if peer.encrypted:
-                message = message.decrypt(peer.key_data.aes_key)
+                try:
+                    message = message.decrypt(peer.key_data.aes_key)
+                except Exception as e:
+                    print(repr(e))
+                    print("can't decrypt")
+                log.debug(f"Received decrypted: {type(message).__name__}")
 
-            if isinstance(message, (PtyMessage, PtyResizeMessage)):
-                peer.handle_pty_message(message)
-                continue
+            match message:
+                case AesRekeyMessage():
+                    await self.handle_aes_rekey_message(peer, message)
 
-            if isinstance(message, FileEntryStreamMessage):
-                await peer.handle_file_stream_message(message)
-                continue
+                case PtyMessage() | PtyResizeMessage():
+                    peer.handle_pty_message(message)
 
-            if isinstance(message, LivelinessMessage):
-                await self.handle_liveliness_message(peer, message)
-                continue
+                case FileEntryStreamMessage():
+                    await peer.handle_file_stream_message(message)
 
-            if isinstance(message, ChallengeReplyMessage):
-                await self.handle_auth_message(peer, message)
-                continue
+                case LivelinessMessage():
+                    await self.handle_liveliness_message(peer, message)
 
-            # Encrypted message is the test message (peer isn't encrypted)
-            if isinstance(message, (SessionKeyMessage, EncryptedMessage)):
-                await self.handle_encryption_message(peer, message)
-                continue
+                case ChallengeReplyMessage():
+                    await self.handle_auth_message(peer, message)
 
-            if isinstance(message, ProxyMessage):
-                await self.handle_forwarding_message(peer, message)
-                continue
+                # Encrypted message is the test message (peer isn't encrypted)
+                case SessionKeyMessage() | EncryptedMessage():
+                    await self.handle_encryption_message(peer, message)
 
-            if isinstance(message, RpcRequestMessage):
-                log.debug(
-                    f"Message received (decrypted and decoded): {bson.decode(message.payload)})"
-                )
-                await self.rpc_messages.put((peer.id, message.payload))
+                case ProxyMessage():
+                    await self.handle_forwarding_message(peer, message)
 
-            else:
-                log.error(f"Unknown message: {message}")
+                case RpcRequestMessage():
+                    log.debug(
+                        f"Message received (decrypted and decoded): {bson.decode(message.payload)})"
+                    )
+                    await self.rpc_messages.put(
+                        (peer.id, message.chan_id, message.payload)
+                    )
+
+                case _:
+                    log.error(f"Unknown message: {message}")
+
+        # we hold a reference to the task in case the read socket loop ends and we have messages
+        # still being processed - we ditch them as the socket is probably toast
+        task = asyncio.current_task()
+        peer.in_flight_messages.remove(task)
 
     async def stop_server(self):
         log.info("Stopping server")
@@ -672,13 +726,13 @@ class EncryptedSocketServerTransport(ServerTransport):
         log.info(f"Serving on {addrs}")
 
     async def receive_message(self) -> tuple:
-        addr, message = await self.rpc_messages.get()
+        addr, channel, message = await self.rpc_messages.get()
         # message = message.as_dict()
-        return addr, message
+        return addr, channel, message
 
-    async def send_reply(self, context: tuple, data: bytes):
-        msg = RpcReplyMessage(data)
-        peer = self.peers.get_peer(context)
+    async def send_reply(self, context: tuple, channel: int, data: bytes):
+        msg = RpcReplyMessage(channel, data)
+        peer: EncryptablePeer = self.peers.get_peer(context)
 
         if not peer:
             # socket has been terminated / removed
